import 'dart:convert';

import 'package:apollopay/src/api/entities/address_entity.dart';
import 'package:apollopay/src/api/entities/entity_entity.dart';
import 'package:apollopay/src/api/entities/payment/customer_entity.dart';
import 'package:apollopay/src/api/entities/payment/invoice_entity.dart';
import 'package:apollopay/src/api/entities/payment/pricing/price_entity.dart';
import 'package:apollopay/src/api/entities/reccurency.dart';
import 'package:mongo_dart/mongo_dart.dart';

enum ChargeSourceType { none, card }

enum ChargeState {
  unknown,
  ready,
  complete,
  declined,
  voided,
  refunded,
  pending
}

List<ChargeState> chargesStatusesToSave = [
  ChargeState.complete,
  ChargeState.voided,
  ChargeState.declined
];

class Charge extends Entity {
  String? id;
  String? token;
  int? rawAmount;
  String? priceToken;
  String? invoiceToken;
  String currency;
  String? source;
  ChargeSourceType sourceType;
  String? client;
  Address? billingAddress;
  Address? shippingAddress;
  String? description;
  String? providerReference;
  DateTime? date;
  DateTime? endDate;
  ChargeState state;
  bool refunded;
  String? subscriptionChargeId;
  Recurrency periodType = Recurrency.onetime;
  int? periodLength;

  // Assigned by the ChargeController
  Price? price;
  // Assigned by the ChargeController
  Invoice? invoice;

  // Getters
  int get amount {
    if (price != null) return price!.amount;

    if (rawAmount != null) return rawAmount!;

    return 0;
  }

  Charge(
      {this.id,
      this.token,
      this.rawAmount,
      this.priceToken,
      this.invoiceToken,
      this.currency = "CAD",
      this.source,
      this.sourceType = ChargeSourceType.card,
      this.client,
      this.billingAddress,
      this.shippingAddress,
      this.description,
      this.providerReference,
      this.state = ChargeState.unknown,
      this.date,
      this.endDate,
      this.refunded = false,
      this.subscriptionChargeId,
      this.periodType = Recurrency.onetime,
      this.periodLength});

  @override
  Map<String, dynamic> toMap({bool databaseFormat = false}) {
    return {
      'token': token,
      'source_type': sourceType.index,
      'amount': rawAmount,
      'price_token': priceToken,
      'invoice_token': invoiceToken,
      'currency': currency,
      'source': source,
      'client': client,
      'billing_address': billingAddress?.toMap(),
      'shipping_address': shippingAddress?.toMap(),
      'description': description,
      if (databaseFormat) 'provider_reference': providerReference,
      'state': state.index,
      'date': date?.millisecondsSinceEpoch,
      'end_date': endDate?.millisecondsSinceEpoch,
      'refunded': refunded,
      'subscription_charge_id': subscriptionChargeId,
      'period_type': periodType.index,
      'period_length': periodLength,
      if (price != null) 'price': price?.toMap(databaseFormat: databaseFormat),
      if (invoice != null) 'invoice': invoice?.toMap(withId: !databaseFormat)
    };
  }

  factory Charge.fromMap(Map<String, dynamic> map) {
    assert(map['source'] != null, 'ChargeError -> source must be set.');
    assert(map['currency'] != null, 'ChargeError -> currency must be set.');
    assert(
        map['amount'] != null ||
            map['price_token'] != null ||
            map['invoice_token'] != null,
        'ChargeError -> amount, price_token or invoice_token must be set.');
    // assert((map['amount'] != null && map['price_token'] == null && map['invoice_token'] == null),
    //     'ChargeError -> must only have amount or price_token.');
    if (map['amount'] != null &&
        map['price_token'] == null &&
        map['invoice_token'] == null) {
      assert(int.parse(map['amount'].toString().replaceAll('.0', '')) >= 100,
          'ChargeError -> amount must be greater than 100 (1\$).');
    }

    return Charge(
        id: map['_id'] is ObjectId
            ? map['_id'].toHexString()
            : map['_id'].toString(),
        sourceType: (map['source_type'] != null)
            ? ChargeSourceType.values[
                int.parse(map['source_type'].toString().replaceAll('.0', ''))]
            : ChargeSourceType.card,
        rawAmount: int.tryParse(map['amount'].toString()) ?? 0,
        priceToken: map['price_token']?.toString(),
        invoiceToken: map['invoice_token']?.toString(),
        currency: map['currency'] ?? '',
        token: map['token'],
        source: map['source'] ?? '',
        client: map['client'],
        billingAddress: map['billing_address'] != null
            ? Address.fromMap(map['billing_address'])
            : null,
        shippingAddress: map['shipping_address'] != null
            ? Address.fromMap(map['shipping_address'])
            : null,
        description: map['description'],
        providerReference: map['provider_reference'],
        state: (map['state'] != null)
            ? ChargeState
                .values[int.parse(map['state'].toString().replaceAll('.0', ''))]
            : ChargeState.unknown,
        date: (map['date'] != null)
            ? DateTime.fromMillisecondsSinceEpoch(
                int.parse(map['date'].toString().replaceAll('.0', '')))
            : DateTime.now(),
        endDate: (map['end_date'] != null)
            ? DateTime.fromMillisecondsSinceEpoch(
                int.parse(map['date'].toString().replaceAll('.0', '')))
            : DateTime.now(),
        refunded: (map['refunded'] != null) ? map['refunded'] : false,
        subscriptionChargeId: map['subscription_charge_id'],
        periodType: map['period_type'] != null
            ? Recurrency.values[int.tryParse(map['period_type'].toString()) ??
                Recurrency.onetime.index]
            : Recurrency.onetime,
        periodLength: map['period_length']);
  }

  factory Charge.fromJson(String source) => Charge.fromMap(json.decode(source));

  @override
  String toString() {
    return 'Payment(amount: $amount, currency: $currency, source: $source, client: $client, billingAddress: $billingAddress, shippingAddress: $shippingAddress, description: $description)';
  }

  List<dynamic> toCSVLine({Customer? customer}) {
    return [
      date?.toLocal(),
      token,
      customer?.toString() ?? "N/A",
      (amount / 100).toDouble().toStringAsFixed(2),
      state.name,
    ];
  }

  @override
  bool operator ==(Object? other) {
    if (identical(this, other)) return true;

    return other is Charge &&
        other.amount == amount &&
        other.currency == currency &&
        other.source == source &&
        other.client == client &&
        other.billingAddress == billingAddress &&
        other.shippingAddress == shippingAddress &&
        other.description == description;
  }

  @override
  int get hashCode {
    return amount.hashCode ^
        currency.hashCode ^
        source.hashCode ^
        client.hashCode ^
        billingAddress.hashCode ^
        shippingAddress.hashCode ^
        description.hashCode;
  }

  @override
  Charge copyWith(
      {Address? billingAddress,
      Address? shippingAddress,
      String? description,
      ChargeState? state,
      bool? refunded}) {
    return Charge(
        id: id,
        token: token,
        rawAmount: amount,
        priceToken: priceToken,
        invoiceToken: invoiceToken,
        currency: currency,
        source: source,
        client: client,
        sourceType: sourceType,
        billingAddress: billingAddress ?? this.billingAddress,
        shippingAddress: shippingAddress ?? this.shippingAddress,
        description: description ?? this.description,
        providerReference: providerReference,
        state: state ?? this.state,
        date: date,
        refunded: refunded ?? this.refunded,
        subscriptionChargeId: subscriptionChargeId,
        periodType: periodType);
  }
}
