import 'dart:convert';
import 'package:apollopay/src/api/config.dart';
import 'package:apollopay/src/api/entities/paginated_entity.dart';
import 'package:apollopay/src/api/entities/payment/invoice_entity.dart';
import 'package:apollopay/src/api/services/client_http_service.dart';
import 'package:http/http.dart' as http;

class InvoicesService extends ClientHttpService {
  InvoicesService({
    required String baseUrl,
    String endpoint = "/pay/invoices/",
  }) : super(baseUrl: baseUrl, endpoint: endpoint);

  Future<Paginated<Invoice>> getInvoices(
      {Map<String, dynamic>? queryParameters}) async {
    String queryParams = "?";
    if (queryParameters != null) {
      queryParameters.forEach((key, value) {
        queryParams += '$key=$value&';
      });
    }
    return await http.get(Uri.parse('$baseUrl$endpoint$queryParams'), headers: {
      'content-type': 'application/json',
      'x-auth-token': applicationToken ?? '',
      'x-auth-private': Config.instance.clientPrivateKey,
      'x-auth-public': Config.instance.clientPublicKey,
    }).then((response) async {
      switch (response.statusCode) {
        case 200:
          Map data = jsonDecode(response.body);
          List<Invoice> invoices = [];
          for (var i = 0; i < data['data'].length; i++) {
            invoices.add(Invoice.fromMap(data['data'][i]));
          }
          var paginated = Paginated<Invoice>.fromJson(data.cast());
          paginated.data = invoices;
          return paginated;
        default:
          throw Exception("Failed to get invoices.");
      }
    });
  }

  Future<Invoice?> getInvoice(String invoiceToken) async {
    return await http.get(
      Uri.parse('$baseUrl$endpoint$invoiceToken'),
      headers: {
        'content-type': 'application/json',
        'x-auth-token': applicationToken ?? '',
        'x-auth-private': Config.instance.clientPrivateKey,
        'x-auth-public': Config.instance.clientPublicKey,
      },
    ).then((response) async {
      switch (response.statusCode) {
        case 200:
          return Invoice.fromJson(response.body);
        default:
          return null;
      }
    });
  }

  Future<Invoice?> getPublicInvoice(
      String clientToken, String invoiceToken) async {
    return await http.get(
      Uri.parse('$baseUrl$endpoint$clientToken/$invoiceToken'),
      headers: {
        'content-type': 'application/json',
        'x-auth-token': applicationToken ?? '',
        'x-auth-private': Config.instance.clientPrivateKey,
        'x-auth-public': Config.instance.clientPublicKey,
      },
    ).then((response) async {
      switch (response.statusCode) {
        case 200:
          return Invoice.fromJson(response.body);
        default:
          return null;
      }
    });
  }

  Future<Invoice?> createInvoice(Invoice invoice) async {
    return await http
        .post(Uri.parse('$baseUrl$endpoint'),
            headers: {
              'content-type': 'application/json',
              'x-auth-token': applicationToken ?? '',
              'x-auth-private': Config.instance.clientPrivateKey,
              'x-auth-public': Config.instance.clientPublicKey,
            },
            body: jsonEncode({"invoice": invoice.toMap()}))
        .then((response) async {
      switch (response.statusCode) {
        case 200:
          return Invoice.fromJson(response.body);
        default:
          return null;
      }
    });
  }

  Future<bool?> modifyInvoice(Invoice invoice) async {
    return await http
        .patch(Uri.parse('$baseUrl$endpoint${invoice.token}'),
            headers: {
              'content-type': 'application/json',
              'x-auth-token': applicationToken ?? '',
              'x-auth-private': Config.instance.clientPrivateKey,
              'x-auth-public': Config.instance.clientPublicKey,
            },
            body: jsonEncode({"invoice": invoice.toMap()}))
        .then((response) async {
      switch (response.statusCode) {
        case 200:
          return true;
        default:
          return null;
      }
    });
  }

  Future<bool?> deleteInvoice(String invoiceToken) async {
    return await http
        .delete(Uri.parse('$baseUrl$endpoint$invoiceToken'), headers: {
      'content-type': 'application/json',
      'x-auth-token': applicationToken ?? '',
      'x-auth-private': Config.instance.clientPrivateKey,
      'x-auth-public': Config.instance.clientPublicKey,
    }).then((response) async {
      switch (response.statusCode) {
        case 200:
          return true;
        default:
          return null;
      }
    });
  }
}
