import 'dart:convert';
import 'package:apollopay/src/api/config.dart';
import 'package:apollopay/src/api/entities/paginated_entity.dart';
import 'package:apollopay/src/api/entities/payment/product_entity.dart';
import 'package:apollopay/src/api/services/client_http_service.dart';
import 'package:http/http.dart' as http;

class ProductsService extends ClientHttpService {
  ProductsService({
    required String baseUrl,
    String endpoint = "/pay/products/",
  }) : super(baseUrl: baseUrl, endpoint: endpoint);

  Future<Paginated<Product>> getProducts(
      {Map<String, dynamic>? queryParameters}) async {
    String queryParams = "?";
    if (queryParameters != null) {
      queryParameters.forEach((key, value) {
        queryParams += '$key=$value&';
      });
    }
    return await http.get(Uri.parse('$baseUrl$endpoint$queryParams'), headers: {
      'content-type': 'application/json',
      'x-auth-token': applicationToken ?? '',
      'x-auth-private': Config.instance.clientPrivateKey,
      'x-auth-public': Config.instance.clientPublicKey,
    }).then((response) async {
      switch (response.statusCode) {
        case 200:
          Map data = jsonDecode(response.body);
          List<Product> products = [];
          for (var i = 0; i < data['data'].length; i++) {
            products.add(Product.fromMap(data['data'][i]));
          }
          var paginated = Paginated<Product>.fromJson(data.cast());
          paginated.data = products;
          return paginated;
        default:
          throw Exception("Failed to get products.");
      }
    });
  }

  Future<Product?> getProduct(String productToken,
      {bool showInactive = false}) async {
    return await http.get(
      Uri.parse('$baseUrl$endpoint$productToken?showInactive=$showInactive'),
      headers: {
        'content-type': 'application/json',
        'x-auth-token': applicationToken ?? '',
        'x-auth-private': Config.instance.clientPrivateKey,
        'x-auth-public': Config.instance.clientPublicKey,
      },
    ).then((response) async {
      switch (response.statusCode) {
        case 200:
          var body = jsonDecode(response.body) as Map<String, dynamic>;
          return Product.fromMap(body);
        default:
          return null;
      }
    });
  }

  Future<Product?> createProduct(Product product) async {
    return await http
        .post(Uri.parse('$baseUrl$endpoint'),
            headers: {
              'content-type': 'application/json',
              'x-auth-token': applicationToken ?? '',
              'x-auth-private': Config.instance.clientPrivateKey,
              'x-auth-public': Config.instance.clientPublicKey,
            },
            body: jsonEncode({"product": product.toMap()}))
        .then((response) async {
      switch (response.statusCode) {
        case 200:
          return Product.fromJson(response.body);
        default:
          return null;
      }
    });
  }

  Future<bool?> modifyProduct(Product product) async {
    return await http
        .patch(Uri.parse('$baseUrl$endpoint${product.token}'),
            headers: {
              'content-type': 'application/json',
              'x-auth-token': applicationToken ?? '',
              'x-auth-private': Config.instance.clientPrivateKey,
              'x-auth-public': Config.instance.clientPublicKey,
            },
            body: jsonEncode({"product": product.toMap()}))
        .then((response) async {
      switch (response.statusCode) {
        case 200:
          return true;
        default:
          return null;
      }
    });
  }

  Future<bool?> deleteProduct(String productToken) async {
    return await http
        .delete(Uri.parse('$baseUrl$endpoint$productToken'), headers: {
      'content-type': 'application/json',
      'x-auth-token': applicationToken ?? '',
      'x-auth-private': Config.instance.clientPrivateKey,
      'x-auth-public': Config.instance.clientPublicKey,
    }).then((response) async {
      switch (response.statusCode) {
        case 200:
          return true;
        default:
          return null;
      }
    });
  }
}
